# -*- coding: utf-8 -*-

"""Download information from several registries."""

import json
import logging
import os
import tempfile
from typing import Any, Iterable, Mapping, Optional
from urllib.request import urlretrieve

import requests
import yaml

__all__ = [
    'get_obofoundry',
    'get_miriam',
    'get_ols',
    'get_curated_registry',
    'get_namespace_synonyms',
    'get_metaregistry',
]

logger = logging.getLogger(__name__)

HERE = os.path.abspath(os.path.dirname(__file__))

OBOFOUNDRY_URL = 'https://raw.githubusercontent.com/OBOFoundry/OBOFoundry.github.io/master/registry/ontologies.yml'
MIRIAM_URL = 'https://registry.api.identifiers.org/restApi/namespaces'
OLS_URL = 'http://www.ebi.ac.uk/ols/api/ontologies'

OBOFOUNDRY_CACHE_PATH = os.path.join(HERE, 'obofoundry.json')
MIRIAM_CACHE_PATH = os.path.join(HERE, 'miriam.json')
OLS_CACHE_PATH = os.path.join(HERE, 'ols.json')

#: The self-curated registry metadatabase
METAREGISTRY_PATH = os.path.join(HERE, 'metaregistry.json')


def get_obofoundry(cache_path: Optional[str] = OBOFOUNDRY_CACHE_PATH, mappify: bool = False):
    """Get the OBO Foundry registry."""
    if cache_path is not None and os.path.exists(cache_path):
        with open(cache_path) as file:
            rv = json.load(file)
            if mappify:
                return _mappify(rv, 'id')
            return rv

    with tempfile.TemporaryDirectory() as d:
        yaml_path = os.path.join(d, 'obofoundry.yml')
        urlretrieve(OBOFOUNDRY_URL, yaml_path)
        with open(yaml_path) as file:
            rv = yaml.full_load(file)

    rv = rv['ontologies']
    for s in rv:
        for k in ('browsers', 'usages', 'depicted_by', 'products'):
            if k in s:
                del s[k]

    if cache_path is not None:
        with open(cache_path, 'w') as file:
            json.dump(rv, file, indent=2)

    if mappify:
        rv = _mappify(rv, 'id')

    return rv


def _mappify(j, k):
    return {entry[k]: entry for entry in j}


def get_miriam(cache_path: Optional[str] = MIRIAM_CACHE_PATH, mappify: bool = False):
    """Get the MIRIAM registry."""
    return _ensure(MIRIAM_URL, embedded_key='namespaces', cache_path=cache_path, mappify_key=mappify and 'prefix')


def get_ols(cache_path: Optional[str] = OLS_CACHE_PATH, mappify: bool = False):
    """Get the OLS registry."""
    return _ensure(OLS_URL, embedded_key='ontologies', cache_path=cache_path, mappify_key=mappify and 'ontologyId')


def _ensure(url, embedded_key, cache_path: Optional[str] = None, mappify_key: Optional[str] = None):
    if cache_path is not None and os.path.exists(cache_path):
        with open(cache_path) as file:
            return json.load(file)

    rv = _download_paginated(url, embedded_key=embedded_key)
    if cache_path is not None:
        with open(cache_path, 'w') as file:
            json.dump(rv, file, indent=2)

    if mappify_key is not None:
        rv = _mappify(rv, mappify_key)

    return rv


def _download_paginated(start_url, embedded_key):
    results = []
    url = start_url
    while True:
        logger.debug('getting', url)
        g = requests.get(url)
        j = g.json()
        r = j['_embedded'][embedded_key]
        results.extend(r)
        links = j['_links']
        if 'next' not in links:
            break
        url = links['next']['href']
    return results


def get_curated_registry():
    """Get the metaregistry."""
    with open(METAREGISTRY_PATH) as file:
        return json.load(file)


def get_namespace_synonyms() -> Mapping[str, str]:
    """Return a mapping from several variants of each synonym to the canonical namespace."""
    synonym_to_key = {}

    def _add_variety(_synonym, _target) -> None:
        synonym_to_key[_synonym] = _target
        synonym_to_key[_synonym.lower()] = _target
        synonym_to_key[_synonym.upper()] = _target
        synonym_to_key[_synonym.casefold()] = _target
        for x, y in [('_', ' '), (' ', '_'), (' ', '')]:
            synonym_to_key[_synonym.replace(x, y)] = _target
            synonym_to_key[_synonym.lower().replace(x, y)] = _target
            synonym_to_key[_synonym.upper().replace(x, y)] = _target
            synonym_to_key[_synonym.casefold().replace(x, y)] = _target

    for entry in get_miriam():
        prefix, name = entry['prefix'], entry['name']
        _add_variety(prefix, prefix)
        _add_variety(name, prefix)

    for entry in get_ols():
        ontology_id = entry['ontologyId']
        _add_variety(ontology_id, ontology_id)
        _add_variety(entry['config']['title'], ontology_id)
        _add_variety(entry['config']['namespace'], ontology_id)

    metaregistry = get_curated_registry()
    for key, values in metaregistry['database'].items():
        _add_variety(key, key)
        for synonym in values.get('synonyms', []):
            _add_variety(synonym, key)

    return synonym_to_key


def get_metaregistry() -> Iterable[Mapping[str, Any]]:
    """Get a combine registry."""
    # TODO!
    for namespace in get_miriam():
        yield dict(
            prefix=namespace['prefix'],
            name=namespace['name'],
            pattern=namespace['pattern'],
            namespace_in_pattern=namespace['namespaceEmbeddedInLui'],
        )

    # for obo in get_converted_obos():
    #     if obo.pattern is not None:
    #         continue
    #     yield dict(
    #         prefix=obo.ontology,
    #         name=obo.name,
    #         pattern=obo.pattern,
    #         namespace_in_pattern=obo.namespace_in_pattern,
    #     )



# -*- coding: utf-8 -*-

"""SQLAlchemy models for OBO."""

from __future__ import annotations

from sqlalchemy import Column, ForeignKey, Integer, String, Table
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, relationship

__all__ = [
    'Base',
    'Reference',
    'Term',
    'term_parents',
]

REFERENCE_TABLE_NAME = 'pyobo_reference'
TERM_TABLE_NAME = 'pyobo_term'
PARENTS_TABLE_NAME = 'pyobo_parents'

Base = declarative_base()

term_parents = Table(
    PARENTS_TABLE_NAME, Base.metadata,
    Column('term_id', Integer, ForeignKey(f'{TERM_TABLE_NAME}.id'), primary_key=True),
    Column('reference_id', Integer, ForeignKey(f'{REFERENCE_TABLE_NAME}.id'), primary_key=True)
)


class Reference(Base):
    """Represent a CURIE and label."""

    __tablename__ = REFERENCE_TABLE_NAME
    id = Column(Integer, primary_key=True)

    namespace = Column(String)
    identifier = Column(String)
    name = Column(String)


class Term(Base):
    """Represent an OBO term."""

    __tablename__ = TERM_TABLE_NAME
    id = Column(Integer, primary_key=True)

    reference_id = Column(Integer, ForeignKey(f'{Reference.__tablename__}.id'), nullable=False, index=True)
    reference = relationship(Reference)

    definition = Column(String)

    parents = relationship(
        Reference,
        secondary=term_parents,
        lazy='dynamic',
        backref=backref('children', lazy='dynamic'),
    )



# -*- coding: utf-8 -*-

"""Utilities for caching files."""

import functools
import json
import logging
import os
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Union

import networkx as nx
import obonet
import pandas as pd

from .io_utils import open_map_tsv, open_multimap_tsv, write_map_tsv, write_multimap_tsv

logger = logging.getLogger(__name__)

JSONType = Union[
    Dict[str, Any],
    List[Any],
]

MappingGetter = Callable[[], Mapping[str, str]]
MultiMappingGetter = Callable[[], Mapping[str, List[str]]]
JSONGetter = Callable[[], JSONType]
GraphGetter = Callable[[], nx.MultiDiGraph]
DataFrameGetter = Callable[[], pd.DataFrame]


def cached_mapping(path: str, header: Iterable[str]) -> Callable[[MappingGetter], MappingGetter]:  # noqa: D202
    """Create a decorator to apply to a mapping getter."""

    def wrapped(f: MappingGetter) -> MappingGetter:  # noqa: D202
        """Wrap a mapping getter so it can be auto-loaded from a cache."""

        @functools.wraps(f)
        def _wrapped() -> Mapping[str, str]:
            if os.path.exists(path):
                return open_map_tsv(path)
            rv = f()
            write_map_tsv(path=path, header=header, rv=rv)
            return rv

        return _wrapped

    return wrapped


def cached_json(path: str) -> Callable[[JSONGetter], JSONGetter]:  # noqa: D202
    """Create a decorator to apply to a mapping getter."""

    def wrapped(f: JSONGetter) -> JSONGetter:  # noqa: D202
        """Wrap a mapping getter so it can be auto-loaded from a cache."""

        @functools.wraps(f)
        def _wrapped() -> JSONType:
            if os.path.exists(path):
                with open(path) as file:
                    return json.load(file)
            rv = f()
            with open(path, 'w') as file:
                json.dump(rv, file, indent=2)
            return rv

        return _wrapped

    return wrapped


def cached_graph(path: str) -> Callable[[GraphGetter], GraphGetter]:  # noqa: D202
    """Create a decorator to apply to a graph getter."""

    def wrapped(f: GraphGetter) -> GraphGetter:  # noqa: D202
        """Wrap a mapping getter so it can be auto-loaded from a cache."""

        @functools.wraps(f)
        def _wrapped() -> nx.MultiDiGraph:
            if os.path.exists(path):
                logger.debug('loading graph from pickle: %s', path)
                return nx.read_gpickle(path)
            rv = f()
            nx.write_gpickle(rv, path)
            return rv

        return _wrapped

    return wrapped


def cached_df(path: str, sep: str = '\t', **kwargs):  # noqa: D202
    """Create a decorator to apply to a dataframe getter."""

    def wrapped(f: DataFrameGetter) -> DataFrameGetter:  # noqa: D202
        """Wrap a mapping getter so it can be auto-loaded from a cache."""

        @functools.wraps(f)
        def _wrapped() -> pd.DataFrame:
            if os.path.exists(path):
                return pd.read_csv(path, sep=sep, **kwargs)
            rv = f()
            rv.to_csv(path, sep=sep, index=False)
            return rv

        return _wrapped

    return wrapped


def cached_multidict(path: str, header: Iterable[str]):  # noqa: D202
    """Create a decorator to apply to a dataframe getter."""

    def wrapped(f: MultiMappingGetter) -> MultiMappingGetter:  # noqa: D202
        """Wrap a mapping getter so it can be auto-loaded from a cache."""

        @functools.wraps(f)
        def _wrapped() -> Mapping[str, List[str]]:
            if os.path.exists(path):
                return open_multimap_tsv(path)
            rv = f()
            write_multimap_tsv(path=path, header=header, rv=rv)
            return rv

        return _wrapped

    return wrapped


def ensure_obo_graph(path: str, pickle_path: Optional[str] = None) -> nx.MultiDiGraph:
    """Get an OBO graph from a given path."""
    if pickle_path is None:
        pickle_path = f'{path}.pickle'

    @cached_graph(path=pickle_path)
    def _read_obo() -> nx.MultiDiGraph:
        return obonet.read_obo(path)

    return _read_obo()



# -*- coding: utf-8 -*-

"""Uploader for OBO."""

import click
from tqdm import tqdm

from pyobo import get_obo_graph, get_terms_from_graph
from pyobo.database.sql.manager import Manager
from pyobo.database.sql.models import Reference, Term


@click.command()
def main():
    """Upload pickle graph."""
    g = get_obo_graph('go')
    terms = get_terms_from_graph(g)

    references = {}

    manager = Manager()
    manager.create_all()

    for term in tqdm(terms[:500]):
        reference = references.get(term.reference.curie)
        if reference is None:
            reference = references[term.reference.curie] = Reference(
                namespace=term.reference.namespace,
                identifier=term.reference.identifier,
                name=term.reference.name
            )

        term = Term(
            reference=reference,
            definition=term.definition,
        )

        manager.session.add(term)

    manager.session.commit()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""CLI for PyOBO."""

import click

from .mappings import get_all_xrefs, get_id_name_mapping, get_synonyms, get_xrefs

__all__ = ['main']


@click.group()
def main():
    """CLI for PyOBO."""


prefix_argument = click.argument('prefix')


@main.command()
@prefix_argument
@click.option('--target')
def xrefs(prefix: str, target: str):
    """Page through xrefs for the given namespace to the second given namespace."""
    if target:
        click.echo_via_pager('\n'.join(
            f'{identifier}\t{_xref}'
            for identifier, _xrefs in get_xrefs(prefix, target).items()
            for _xref in _xrefs
        ))
    else:
        click.echo_via_pager('\n'.join(
            '\t'.join(row)
            for row in get_all_xrefs(prefix).values
        ))


@main.command()
@prefix_argument
def names(prefix: str):
    """Page through the identifiers and names of entities in the given namespace."""
    click.echo_via_pager('\n'.join(
        '\t'.join(item)
        for item in get_id_name_mapping(prefix).items()
    ))


@main.command()
@prefix_argument
def synonyms(prefix: str):
    """Page through the synonyms for entities in the given namespace."""
    click.echo_via_pager('\n'.join(
        f'{identifier}\t{_synonym}'
        for identifier, _synonyms in get_synonyms(prefix).items()
        for _synonym in _synonyms
    ))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Constants for PyOBO."""

import os

__all__ = [
    'PYOBO_HOME',
    'OUTPUT_DIRECTORY',
    'CURATED_URLS',
]

PYOBO_HOME = os.environ.get('PYOBO_HOME') or os.path.join(os.path.expanduser('~'), '.obo')

OUTPUT_DIRECTORY = (
    os.environ.get('PYOBO_OUTPUT')
    or os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, os.pardir, 'obo'))
)

#: A curated list of prefixes to URLs for OBO files that aren't properly listed in OBO Foundry
CURATED_URLS = {
    'mp': 'http://purl.obolibrary.org/obo/mp.obo',
    'chiro': 'http://purl.obolibrary.org/obo/chiro.obo',
}



# -*- coding: utf-8 -*-

"""This module contains the base class for connection managers in SQLAlchemy."""

import logging
from typing import Tuple

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from .models import Base

__all__ = [
    'Manager',
    'build_engine_session',
]

log = logging.getLogger(__name__)


def build_engine_session() -> Tuple:
    """Build a session."""
    engine = create_engine('sqlite:////Users/cthoyt/Desktop/obo.db')

    #: A SQLAlchemy session maker
    session_maker = sessionmaker(bind=engine)

    #: A SQLAlchemy session object
    session = scoped_session(session_maker)

    return engine, session


class Manager:
    """A wrapper around a SQLAlchemy engine and session."""

    #: The declarative base for this manager
    base = Base

    def __init__(self, engine=None, session=None) -> None:
        """Instantiate a manager from an engine and session."""
        if engine is None and session is None:
            engine, session = build_engine_session()
        self.engine = engine
        self.session = session

    def create_all(self, checkfirst: bool = True) -> None:
        """Create the PyBEL cache's database and tables.

        :param checkfirst: Check if the database exists before trying to re-make it
        """
        self.base.metadata.create_all(bind=self.engine, checkfirst=checkfirst)

    def drop_all(self, checkfirst: bool = True) -> None:
        """Drop all data, tables, and databases for the PyBEL cache.

        :param checkfirst: Check if the database exists before trying to drop it
        """
        self.session.close()
        self.base.metadata.drop_all(bind=self.engine, checkfirst=checkfirst)

    def bind(self) -> None:
        """Bind the metadata to the engine and session."""
        self.base.metadata.bind = self.engine
        self.base.query = self.session.query_property()

    def __repr__(self):  # noqa: D105
        return '<{} connection={}>'.format(self.__class__.__name__, self.engine.url)



# -*- coding: utf-8 -*-

"""Extract *all* xrefs from OBO documents available."""

import os
from typing import Iterable, Set, Tuple

import click
import pandas as pd

from .extract_xrefs import UNHANDLED_NAMESPACES, iterate_xrefs_from_graph
from ..getters import MissingOboBuild, get_obo_graph_by_prefix, get_obo_graph_by_url

#: Keys are prefixes and values point to OBO URLs to download
OBO = {
    # High quality
    'hp': 'http://purl.obolibrary.org/obo/hp.obo',
    "chebi": "http://purl.obolibrary.org/obo/chebi.obo",
    "chiro": "https://raw.githubusercontent.com/obophenotype/chiro/master/chiro.obo",
    'doid': 'http://purl.obolibrary.org/obo/doid.obo',
    'efo': 'http://www.ebi.ac.uk/efo/efo.obo',
    "go": "http://purl.obolibrary.org/obo/go.obo",
    "obi": "http://purl.obolibrary.org/obo/obi.obo",
    # Others
    "pr": "http://purl.obolibrary.org/obo/pr.obo",
    "bto": "http://purl.obolibrary.org/obo/bto.obo",
    "cl": "http://purl.obolibrary.org/obo/cl.obo",
    # "clo":  # not distributed as OBO
    "cmo": "http://purl.obolibrary.org/obo/cmo.obo",
    "ecto": "http://purl.obolibrary.org/obo/ecto.obo",
    "exo": "http://purl.obolibrary.org/obo/exo.obo",
    "fbbt": "http://purl.obolibrary.org/obo/fbbt.obo",
    'mondo': 'http://purl.obolibrary.org/obo/mondo.obo',
    "mp": "http://purl.obolibrary.org/obo/mp.obo",
    "mpath": "https://raw.githubusercontent.com/PaulNSchofield/mpath/master/mpath.obo",
    'ncit': 'http://purl.obolibrary.org/obo/ncit.obo',
    "pato": "http://purl.obolibrary.org/obo/pato.obo",
    "peco": "http://purl.obolibrary.org/obo/peco.obo",
    "pw": "http://purl.obolibrary.org/obo/pw.obo",
    'symp': 'http://purl.obolibrary.org/obo/symp.obo',
    "to": "http://purl.obolibrary.org/obo/to.obo",
    "uberon": "http://purl.obolibrary.org/obo/uberon/basic.obo",
}

AUTOGENERATED = {
    "hgnc": None,
    "mirbase": None,
    "sgd": None,
    "hgnc.genefamily": None,
    "mgi": None,
    "rgd": None,
}
OBO.update(AUTOGENERATED)


def _iterate_all_xrefs() -> Iterable[Tuple[str, str, str, str, str]]:
    for prefix, url in OBO.items():
        try:
            graph = get_obo_graph_by_prefix(prefix)
        except MissingOboBuild:
            graph = get_obo_graph_by_url(prefix, url=url)

        for head_ns, head_id, xref_ns, xref_id in iterate_xrefs_from_graph(graph):
            yield head_ns, head_id, xref_ns, xref_id, prefix


def _get_xrefs_df() -> pd.DataFrame:
    xrefs: Set[Tuple[str, str, str, str, str]] = set(_iterate_all_xrefs())
    columns = ['head_ns', 'head_id', 'tail_ns', 'tail_id', 'source']
    return pd.DataFrame(xrefs, columns=columns).sort_values(columns)


@click.command()
@click.option('--directory', type=click.Path(dir_okay=True, file_okay=False, exists=True), default=os.getcwd())
def main(directory):  # noqa: D202
    """Output the mappings."""

    def _write_tsv(df: pd.DataFrame, name: str) -> None:
        df.to_csv(os.path.join(directory, name), sep='\t', index=False)

    xrefs_df = _get_xrefs_df()

    # Export all xrefs
    _write_tsv(xrefs_df, f'xrefs.tsv')
    _write_tsv(xrefs_df, f'xrefs.tsv.gz')

    # Export a sample of xrefs
    _write_tsv(xrefs_df.head(), f'xrefs_sample.tsv')

    # Export a summary dataframe
    summary_df = xrefs_df.groupby(['source', 'tail_ns'])['head_ns'].count().reset_index()
    summary_df = summary_df.sort_values(['head_ns'], ascending=False)
    _write_tsv(summary_df, 'summary.tsv')

    # Export the namespaces that haven't been handled yet
    unmapped_path = os.path.join(directory, 'unmapped.tsv')
    with open(unmapped_path, 'w') as file:
        for namespace, items in sorted(UNHANDLED_NAMESPACES.items()):
            for curie, xref in items:
                print(curie, namespace, xref, file=file, sep='\t')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""This script compares what's in OBO, OLS, and MIRIAM."""

import json
import os

import matplotlib.pyplot as plt
from matplotlib_venn import venn3

HERE = os.path.abspath(os.path.dirname(__file__))


def _get_json(name):
    with open(os.path.join(HERE, name)) as file:
        return json.load(file)


def main():
    """Compare the registries."""
    miriam = _get_json('miriam.json')
    miriam_entries = {
        entry['prefix'].lower()
        for entry in miriam
    }

    ols = _get_json('ols.json')
    ols_entries = {
        entry['ontologyId'].lower()
        for entry in ols
    }

    obofoundry = _get_json('obofoundry.json')
    obofoundry_entries = {
        entry['id'].lower()
        for entry in obofoundry
    }

    venn3(
        subsets=[miriam_entries, ols_entries, obofoundry_entries],
        set_labels=('MIRIAM', 'OLS', 'OBOFoundry'),
    )
    plt.savefig('compare.png')

    # nothing interesting unique to OLS
    # print(*sorted(ols_entries - miriam_entries - obofoundry_entries), sep='\n')
    # nothing interesting unique in OBO
    # print(*sorted(obofoundry_entries - miriam_entries - ols_entries), sep='\n')
    # Some things missing from miriam
    # print(*sorted(obofoundry_entries.union(ols_entries) - miriam_entries), sep='\n')

    # Stuff important enough to make it everywhere
    print(*sorted(set.intersection(ols_entries, miriam_entries, obofoundry_entries)), sep='\n')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Converter for RGD."""

from typing import Iterable

import pandas as pd
from tqdm import tqdm

from ..path_utils import ensure_df
from ..struct import Obo, Reference, Synonym, SynonymTypeDef, Term

PREFIX = 'rgd'

old_symbol_type = SynonymTypeDef(id='old_symbol', name='old symbol')
old_name_type = SynonymTypeDef(id='old_name', name='old name')

GENES_URL = 'ftp://ftp.rgd.mcw.edu/pub/data_release/GENES_RAT.txt'
GENES_HEADER = [
    'GENE_RGD_ID',
    'SYMBOL',
    'NAME',
    'GENE_DESC',
    'CHROMOSOME_CELERA',
    'CHROMOSOME_[oldAssembly#] chromosome for the old reference assembly',
    'CHROMOSOME_[newAssembly#] chromosome for the current reference assembly',
    'FISH_BAND',
    'START_POS_CELERA',
    'STOP_POS_CELERA',
    'STRAND_CELERA',
    'START_POS_[oldAssembly#]',
    'STOP_POS_[oldAssembly#]',
    'STRAND_[oldAssembly#]',
    'START_POS_[newAssembly#]',
    'STOP_POS_[newAssembly#]',
    'STRAND_[newAssembly#]',
    'CURATED_REF_RGD_ID',
    'CURATED_REF_PUBMED_ID',
    'UNCURATED_PUBMED_ID',
    'NCBI_GENE_ID',
    'UNIPROT_ID',
    'UNCURATED_REF_MEDLINE_ID',
    'GENBANK_NUCLEOTIDE',
    'TIGR_ID',
    'GENBANK_PROTEIN',
    'UNIGENE_ID',
    'SSLP_RGD_ID',
    'SSLP_SYMBOL',
    'OLD_SYMBOL',
    'OLD_NAME',
    'QTL_RGD_ID',
    'QTL_SYMBOL',
    'NOMENCLATURE_STATUS',
    'SPLICE_RGD_ID',
    'SPLICE_SYMBOL',
    'GENE_TYPE',
    'ENSEMBL_ID',
]


def get_obo() -> Obo:
    """Get RGD as OBO."""
    terms = list(get_terms())
    return Obo(
        ontology=PREFIX,
        name='Rat Genome Database',
        terms=terms,
        synonym_typedefs=[old_name_type, old_symbol_type]
    )


namespace_to_column = [
    ('ensembl', 'ENSEMBL_ID'),
    ('uniprot', 'UNIPROT_ID'),
    ('unigene', 'UNIGENE_ID'),
    ('ncbigene', 'NCBI_GENE_ID'),
]


def get_terms() -> Iterable[Term]:
    """Get RGD terms."""
    df = ensure_df(PREFIX, GENES_URL, sep='\t', header=0, comment='#', dtype={
        'NCBI_GENE_ID': str,
        'GENE_RGD_ID': str,
    })
    for _, row in tqdm(df.iterrows(), total=len(df.index), desc=f'Mapping {PREFIX}'):
        synonyms = []

        old_names = row['OLD_NAME']
        if old_names and pd.notna(old_names):
            for old_name in old_names.split(';'):
                synonyms.append(Synonym(name=old_name, type=old_name_type))
        old_symbols = row['OLD_SYMBOL']
        if old_symbols and pd.notna(old_symbols):
            for old_symbol in old_symbols.split(';'):
                synonyms.append(Synonym(name=old_symbol, type=old_symbol_type))

        xrefs = []
        for prefix, key in namespace_to_column:
            xref_ids = str(row[key])
            if xref_ids and pd.notna(xref_ids):
                for xref_id in xref_ids.split(';'):
                    xrefs.append(Reference(prefix=prefix, identifier=xref_id))

        provenance = []
        pubmed_ids = row['CURATED_REF_PUBMED_ID']
        if pubmed_ids and pd.notna(pubmed_ids):
            for pubmed_id in str(pubmed_ids).split(';'):
                provenance.append(Reference(prefix='pubmed', identifier=pubmed_id))

        term = Term(
            reference=Reference(prefix=PREFIX, identifier=row['GENE_RGD_ID']),
            name=row['SYMBOL'],
            definition=row['NAME'] or row['GENE_DESC'],
            synonyms=synonyms,
            xrefs=xrefs,
            provenance=provenance,
        )
        yield term


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Converter for Entrez."""

import logging
from typing import Iterable

import pandas as pd
from tqdm import tqdm

from ..mappings import get_id_name_mapping
from ..path_utils import ensure_df
from ..struct import Obo, Reference, Term, from_species

logger = logging.getLogger(__name__)

PREFIX = 'entrez'

SPECIES_CONSORTIUM_MAPPING = {
    '10090': 'mgi',  # Mouse
    '10116': 'rgd',  # Rat
    '4932': 'sgd',  # Yeast
    '7227': 'flybase',  # Drosophila
    '9606': 'hgnc',  # Human
    '6239': 'wormbase',  # C. Elegans
    '7955': 'zfin',  # Zebrafish
}

#: All namepace codes (in lowercase) that can map to ncbigene
CONSORTIUM_SPECIES_MAPPING = {
    database_code: taxonomy_id
    for taxonomy_id, database_code in SPECIES_CONSORTIUM_MAPPING.items()
}

GENE_INFO_URL = 'ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/gene_info.gz'
#: Columns fro gene_info.gz that are used
GENE_INFO_COLUMNS = [
    '#tax_id',
    'GeneID',
    'Symbol',
    'dbXrefs',
    'description',
    'type_of_gene',
]


def get_obo() -> Obo:
    """Get Entrez as OBO."""
    terms = list(get_terms())
    return Obo(
        ontology=PREFIX,
        name='Entrez Gene',
        terms=terms,
        typedefs=[from_species],
        auto_generated_by='bio2obo:entrez',
    )


def get_gene_info_df() -> pd.DataFrame:
    """Get the gene info dataframe."""
    return ensure_df(
        PREFIX,
        GENE_INFO_URL,
        sep='\t',
        na_values=['-', 'NEWENTRY'],
        usecols=GENE_INFO_COLUMNS,
        dtype={'#tax_id': str, 'GeneID': str},
    )


"""xref_mapping was obtained from:

namespaces = set()
for xrefs in df[df['dbXrefs'].notna()]['dbXrefs']:
    for xref in xrefs.split('|'):
        namespaces.add(xref.split(':')[0])

print('namespaces:')
print(*sorted(namespaces), sep='\n')
"""

xref_mapping = {
    'APHIDBASE',
    'ASAP',
    'AnimalQTLdb',
    'ApiDB_CryptoDB',
    'Araport',
    'BEEBASE',
    'BEETLEBASE',
    'BGD',
    'CGNC',
    'ECOCYC',
    'EcoGene',
    'Ensembl',
    'FLYBASE',
    'HGNC',
    'IMGT/GENE-DB',
    'MGI',
    'MIM',
    'NASONIABASE',
    'PseudoCap',
    'RGD',
    'SGD',
    'TAIR',
    'VGNC',
    'VectorBase',
    'WormBase',
    'Xenbase',
    'ZFIN',
    'dictyBase',
    'miRBase',
}
xref_mapping = {x.lower() for x in xref_mapping}


def get_terms() -> Iterable[Term]:
    """Get Entrez terms."""
    df = get_gene_info_df()

    taxonomy_id_to_name = get_id_name_mapping('ncbitaxon')
    it = tqdm(df.values, total=len(df.index), desc=f'mapping {PREFIX}')

    for tax_id, gene_id, symbol, dbxrfs, description, gene_type in it:
        if pd.isna(symbol):
            continue
        try:
            tax_name = taxonomy_id_to_name[tax_id]
        except KeyError:
            logger.warning(f'Could not look up tax_id={tax_id}')
            tax_name = None

        xrefs = []
        if pd.notna(dbxrfs):
            for xref in dbxrfs.split('|'):
                xref_ns, xref_id = xref.split(':', 1)
                xref_ns = xref_ns.lower()
                xrefs.append(Reference(prefix=xref_ns, identifier=xref_id))

        term = Term(
            name=symbol,
            reference=Reference(prefix=PREFIX, identifier=gene_id, name=symbol),
            definition=description,
            xrefs=xrefs,
        )
        term.append_relationship(from_species, Reference(prefix='taxonomy', identifier=tax_id, name=tax_name))
        yield term


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Converter for SGD."""

from urllib.parse import unquote_plus

from ..path_utils import ensure_tar_df
from ..struct import Obo, Reference, Synonym, SynonymTypeDef, Term, from_species

HEADER = ['chromosome', 'database', 'feature', 'start', 'end', 'a', 'b', 'c', 'data']
PREFIX = 'sgd'

URL = 'https://downloads.yeastgenome.org/sequence/' \
      'S288C_reference/genome_releases/S288C_reference_genome_R64-2-1_20150113.tgz'
INNER_PATH = 'S288C_reference_genome_R64-2-1_20150113/saccharomyces_cerevisiae_R64-2-1_20150113.gff'

alias_type = SynonymTypeDef(id='alias', name='alias')


def get_obo() -> Obo:
    """Get SGD as OBO."""
    terms = list(get_terms())
    return Obo(
        ontology=PREFIX,
        name='Saccharomyces Genome Database',
        terms=terms,
        synonym_typedefs=[alias_type],
    )


def get_terms() -> str:
    """Get SGD terms."""
    df = ensure_tar_df(
        prefix=PREFIX,
        url=URL,
        inner_path=INNER_PATH,
        sep='\t',
        skiprows=18,
        header=None,
        names=HEADER,
    )
    df = df[df['feature'] == 'gene']
    for data in df['data']:
        d = dict(entry.split('=') for entry in data.split(';'))

        identifier = d['dbxref'][len('SGD:'):]
        name = d['Name']
        definition = unquote_plus(d['Note'])

        synonyms = []

        aliases = d.get('Alias')
        if aliases:
            for alias in aliases.split(','):
                synonyms.append(Synonym(name=unquote_plus(alias), type=alias_type))

        term = Term(
            reference=Reference(prefix=PREFIX, identifier=identifier),
            name=name,
            definition=definition,
            synonyms=synonyms,
        )
        term.append_relationship(
            from_species,
            Reference(prefix='taxonomy', identifier='4932', name='Saccharomyces cerevisiae'),
        )
        yield term


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Utilities for data structures for OBO."""

OBO_ESCAPE = {
    c: f'\\{c}'
    for c in ':,"\\()[]{}'
}
OBO_ESCAPE[' '] = '\\W'


def obo_escape(string: str) -> str:
    """Escape all funny characters for OBO."""
    return ''.join(
        OBO_ESCAPE.get(character, character)
        for character in string
    )


def comma_separate(elements) -> str:
    """Map a list to strings and make comma separated."""
    return ', '.join(map(str, elements))



# -*- coding: utf-8 -*-

"""Utilities for building paths."""

import logging
import os
import tarfile
from typing import Optional
from urllib.parse import urlparse
from urllib.request import urlretrieve

import pandas as pd

from .constants import PYOBO_HOME

__all__ = [
    'get_prefix_directory',
    'prefix_directory_join',
    'get_prefix_obo_path',
    'ensure_path',
    'ensure_df',
    'ensure_tar_df',
]

logger = logging.getLogger(__name__)


def get_prefix_directory(prefix: str) -> str:
    """Get the directory."""
    directory = os.path.abspath(os.path.join(PYOBO_HOME, prefix))
    os.makedirs(directory, exist_ok=True)
    return directory


def prefix_directory_join(prefix: str, *parts: str) -> str:
    """Join the parts onto the prefix directory."""
    return os.path.join(get_prefix_directory(prefix), *parts)


def get_prefix_obo_path(prefix: str) -> str:
    """Get the canonical path to the OBO file."""
    return prefix_directory_join(prefix, f"{prefix}.obo")


def ensure_path(prefix: str, url: str, path: Optional[str] = None) -> str:
    """Download a file if it doesn't exist."""
    if path is None:
        parse_result = urlparse(url)
        path = os.path.basename(parse_result.path)

    path = prefix_directory_join(prefix, path)

    if not os.path.exists(path):
        logger.info('downloading %s OBO from %s', prefix, url)
        urlretrieve(url, path)

    return path


def ensure_df(prefix: str, url: str, path: Optional[str] = None, **kwargs) -> pd.DataFrame:
    """Download a file and open as a dataframe."""
    path = ensure_path(prefix, url, path=path)
    return pd.read_csv(path, **kwargs)


def ensure_tar_df(prefix: str, url: str, inner_path: str, **kwargs) -> pd.DataFrame:
    """Download a tar file and open as a dataframe."""
    path = ensure_path(prefix, url)
    with tarfile.open(path) as tar_file:
        with tar_file.extractfile(inner_path) as file:
            return pd.read_csv(file, **kwargs)



# -*- coding: utf-8 -*-

"""Sources of OBO content."""

import os
from importlib import import_module
from typing import Iterable

from ..struct import Obo

__all__ = [
    'CONVERTED',
    'get_converted_obo',
    'get_converted_obos',
]

HERE = os.path.abspath(os.path.dirname(__file__))

CONVERTED = {
    'cgnc': 'cgnc',
    'ncbigene': 'entrez',
    'ec-code': 'expasy',
    'hgnc': 'hgnc',
    'hgnc.genefamily': 'hgncgenefamily',
    'mesh': 'mesh',
    'mgi': 'mgi',
    'mirbase': 'mirbase',
    'rgd': 'rgd',
    'sgd': 'sgd',
}


def get_converted_obo(prefix: str) -> Obo:
    """Get a converted PyOBO source."""
    module = import_module(f'pyobo.sources.{CONVERTED[prefix]}')
    return module.get_obo()


def get_converted_obos() -> Iterable[Obo]:
    """Get all modules in the PyOBO sources."""
    for prefix in sorted(CONVERTED):
        yield get_converted_obo(prefix)



# -*- coding: utf-8 -*-

"""Data structures for OBO."""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from operator import attrgetter
from pathlib import Path
from typing import Dict, Iterable, List, Optional, TextIO, Union

from networkx.utils import open_file

from .registry import Registry, miriam
from .utils import comma_separate, obo_escape
from ..path_utils import get_prefix_obo_path

__all__ = [
    'Reference',
    'Synonym',
    'SynonymTypeDef',
    'TypeDef',
    'Term',
    'Obo',
    'from_species',
]


@dataclass
class Reference:
    """A namespace, identifier, and label."""

    #: The namespace's keyword
    prefix: str

    #: The entity's identifier in the namespace
    identifier: str

    name: Optional[str] = field(default=None, repr=None)

    #: The namespace's identifier in the registry
    registry_id: Optional[str] = field(default=None, repr=False)

    #: The registry in which the namespace can be looked up
    registry: Registry = field(default=miriam, repr=False)

    @property
    def curie(self) -> str:  # noqa: D401
        """The CURIE for this reference."""
        return f'{self.prefix}:{self.identifier}'

    @staticmethod
    def from_curie(curie: str) -> Reference:
        """Get a reference from a CURIE."""
        prefix, identifier = curie.strip().split(':')
        return Reference(prefix=prefix, identifier=identifier)

    @staticmethod
    def from_curies(curies: str) -> List[Reference]:
        """Get a list of references from a string with comma separated CURIEs."""
        return [
            Reference.from_curie(curie)
            for curie in curies.split(',')
            if curie.strip()
        ]

    @property
    def _escaped_identifier(self):
        return obo_escape(self.identifier)

    def __str__(self):  # noqa: D105
        if self.identifier.lower().startswith(f'{self.prefix.lower()}:'):
            rv = self.identifier.lower()
        else:
            rv = f'{self.prefix}:{self._escaped_identifier}'
        if self.name:
            rv = f'{rv} ! {self.name}'
        return rv


@dataclass
class Synonym:
    """A synonym with optional specificity and references."""

    #: The string representing the synonym
    name: str

    #: The specificity of the synonym
    specificity: str = 'EXACT'

    #: The type of synonym. Must be defined in OBO document!
    type: Optional[SynonymTypeDef] = None

    #: References to articles where the synonym appears
    provenance: List[Reference] = field(default_factory=list)

    def to_obo(self) -> str:
        """Write this synonym as an OBO line to appear in a [Term] stanza."""
        x = f'synonym: "{self.name}" {self.specificity}'
        if self.type:
            x = f'{x} {self.type.id}'
        return f'{x} [{comma_separate(self.provenance)}]'


@dataclass
class SynonymTypeDef:
    """A type definition for synonyms in OBO."""

    id: str
    name: str

    def to_obo(self) -> str:
        """Serialize to OBO."""
        return f'synonymtypedef: {self.id} "{self.name}"'


@dataclass
class TypeDef:
    """A type definition in OBO."""

    id: str
    name: str
    comment: Optional[str] = None
    namespace: Optional[str] = None
    is_transitive: Optional[bool] = None
    xrefs: List[Reference] = field(default_factory=list)

    def iterate_obo_lines(self) -> Iterable[str]:
        """Iterate over the lines to write in an OBO file."""
        yield '\n[Typedef]'
        yield f'id: {self.id}'
        yield f'name: {self.name}'

        if self.namespace:
            yield f'namespace: {self.namespace}'

        if self.comment:
            yield f'comment: {self.comment}'

        for xref in self.xrefs:
            yield f'xref: {xref}'

        if self.is_transitive is not None:
            yield f'is_transitive: {"true" if self.is_transitive else "false"}'

    def to_obo(self) -> str:
        """Get the OBO document string."""
        return '\n'.join(self.iterate_obo_lines())


@dataclass
class Term:
    """A term in OBO."""

    #: The primary reference for the entity
    reference: Reference

    #: A description of the entity
    definition: Optional[str] = None

    #: References to articles in which the term appears
    provenance: List[Reference] = field(default_factory=list)

    #: Relationships defined by [Typedef] stanzas
    relationships: Dict[str, List[Reference]] = field(default_factory=lambda: defaultdict(list))

    #: Relationships with the default "is_a"
    parents: List[Reference] = field(default_factory=list)

    #: Synonyms of this term
    synonyms: List[Synonym] = field(default_factory=list)

    #: Equivalent references
    xrefs: List[Reference] = field(default_factory=list)

    name: Optional[str] = None

    #: The sub-namespace within the ontology
    namespace: Optional[str] = None

    #: An annotation for obsolescence. By default, is None, but this means that it is not obsolete.
    is_obsolete: Optional[bool] = None

    def get_relationships(self, type_def: TypeDef) -> List[Reference]:
        """Get relationships from the given type."""
        return self.relationships[type_def.id]

    def append_relationship(self, type_def: TypeDef, reference: Reference) -> None:
        """Append a relationship."""
        self.relationships[type_def.id].append(reference)

    def extend_relationship(self, type_def: TypeDef, references: Iterable[Reference]) -> None:
        """Append several relationships."""
        self.relationships[type_def.id].extend(references)

    @property
    def identifier(self) -> str:  # noqa: D401
        """The local unique identifier for this term."""
        return self.reference.identifier

    @property
    def curie(self) -> str:  # noqa: D401
        """The CURIE for this term."""
        return self.reference.curie

    def iterate_obo_lines(self) -> Iterable[str]:
        """Iterate over the lines to write in an OBO file."""
        yield '\n[Term]'
        yield f'id: {self.curie}'
        yield f'name: {self.name}'
        if self.namespace and self.namespace != '?':
            namespace_normalized = self.namespace \
                .replace(' ', '_') \
                .replace('-', '_') \
                .replace('(', '') \
                .replace(')', '')
            yield f'namespace: {namespace_normalized}'

        yield f'''def: "{self.definition}" [{comma_separate(self.provenance)}]'''

        for xref in sorted(self.xrefs, key=attrgetter('prefix', 'identifier')):
            yield f'xref: {xref}'

        for parent in sorted(self.parents, key=attrgetter('prefix', 'identifier')):
            yield f'is_a: {parent}'

        for relationship, relationship_references in sorted(self.relationships.items()):
            for relationship_reference in relationship_references:
                yield f'relationship: {relationship} {relationship_reference}'

        for synonym in sorted(self.synonyms, key=attrgetter('name')):
            yield synonym.to_obo()

    def to_obo(self) -> str:
        """Get the OBO document string."""
        return '\n'.join(self.iterate_obo_lines())


@dataclass
class Obo:
    """An OBO document."""

    #: The prefix for the ontology
    ontology: str

    #: The name of the ontology
    name: str

    #: Terms
    terms: List[Term]

    #: The OBO format
    format_version: str = '1.2'

    #: Type definitions
    typedefs: List[TypeDef] = field(default_factory=list)

    #: Synonym type definitions
    synonym_typedefs: List[SynonymTypeDef] = field(default_factory=list)

    #: Regular expression pattern describing the local unique identifiers
    pattern: Optional[str] = None

    #: Is the prefix at the begging of each local unique identifier
    namespace_in_pattern: Optional[bool] = None

    #: The ontology version
    data_version: Optional[str] = None

    #: An annotation about how an ontology was generated
    auto_generated_by: Optional[str] = None

    #: The date the ontology was generated
    date: datetime = field(default_factory=datetime.today)

    def iterate_obo_lines(self) -> Iterable[str]:
        """Iterate over the lines to write in an OBO file."""
        yield f'format-version: {self.format_version}'
        yield f'date: {self.date.strftime("%d:%m:%Y %H:%M")}'

        if self.auto_generated_by is not None:
            yield f'auto-generated-by: {self.auto_generated_by}'

        if self.data_version is not None:
            yield f'data-version: {self.data_version}'

        for synonym_typedef in sorted(self.synonym_typedefs, key=attrgetter('id')):
            yield synonym_typedef.to_obo()

        yield f'ontology: {self.ontology}'

        for typedef in self.typedefs:
            yield from typedef.iterate_obo_lines()

        for term in self.terms:
            yield from term.iterate_obo_lines()

    def to_obo(self) -> str:
        """Get the OBO document string."""
        return '\n'.join(self.iterate_obo_lines())

    @open_file(1, mode='w')
    def write(self, file: Union[None, str, TextIO, Path] = None) -> None:
        """Write the OBO to a file."""
        print(self.to_obo(), file=file)

    def write_default(self) -> None:
        """Write the OBO to the default path."""
        path = get_prefix_obo_path(self.ontology)
        self.write(path)

    def __iter__(self):  # noqa: D105
        return iter(self.terms)


from_species = TypeDef(
    id='from_species',
    name='from species',
)



# -*- coding: utf-8 -*-

"""Parser for the MeSH descriptors."""

import logging
from typing import Any, Dict, Iterable, List, Mapping, Optional
from xml.etree.ElementTree import Element

from tqdm import tqdm

from ..cache_utils import cached_json, cached_mapping
from ..io_utils import parse_xml_gz
from ..path_utils import ensure_path, prefix_directory_join
from ..struct import Obo, Reference, Synonym, Term

logger = logging.getLogger(__name__)

PREFIX = 'mesh'
YEAR = '2019'
DESCRIPTOR_URL = f'ftp://nlmpubs.nlm.nih.gov/online/mesh/{YEAR}/xmlmesh/desc{YEAR}.gz'


def get_obo() -> Obo:
    """Get MeSH as OBO."""
    terms = list(get_terms())
    return Obo(
        ontology=PREFIX,
        name='Medical Subject Headings',
        terms=terms,
        data_version=YEAR,
    )


@cached_mapping(
    path=prefix_directory_join(PREFIX, f'mesh_{YEAR}_tree.tsv'),
    header=['mesh_tree_number', 'mesh_id'],
)
def get_tree_to_mesh_id() -> Mapping[str, str]:
    """Get a mapping from MeSH tree numbers to their MeSH identifiers."""
    mesh = ensure_mesh()
    rv = {}
    for entry in mesh:
        mesh_id = entry['descriptor_ui']
        for tree_number in entry['tree_numbers']:
            rv[tree_number] = mesh_id
    return rv


def get_terms() -> Iterable[Term]:
    """Get MeSH OBO terms."""
    mesh = ensure_mesh()
    mesh_id_to_term: Dict[str, Term] = {}
    for entry in mesh:
        identifier = entry['descriptor_ui']
        name = entry['name']
        definition = (get_scope_note(entry) or '').strip()

        synonyms = set()
        for concept in entry['concepts']:
            synonyms.add(concept['name'])
            for term in concept['terms']:
                synonyms.add(term['name'])
        synonyms = [
            Synonym(name=synonym)
            for synonym in synonyms
            if synonym != name
        ]

        mesh_id_to_term[identifier] = Term(
            name=name,
            definition=definition,
            reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
            synonyms=synonyms,
        )

    for entry in mesh:
        mesh_id_to_term[entry['descriptor_ui']].parents = [
            mesh_id_to_term[parent_descriptor_id].reference
            for parent_descriptor_id in entry['parents']
        ]

    return mesh_id_to_term.values()


@cached_json(path=prefix_directory_join(PREFIX, f'mesh_{YEAR}.json'))
def ensure_mesh() -> List[Mapping[str, Any]]:
    """Get the parsed MeSH dictionary, and cache it if it wasn't already."""
    path = ensure_path(PREFIX, DESCRIPTOR_URL)
    root = parse_xml_gz(path)
    return get_descriptor_records(root)


def get_descriptor_records(element: Element) -> List[Mapping]:
    """Get MeSH descriptor records."""
    logger.info('extract MeSH descriptors, concepts, and terms')

    rv = [
        get_descriptor_record(descriptor)
        for descriptor in tqdm(element, desc='Getting MeSH Descriptors')
    ]
    logger.debug(f'got {len(rv)} descriptors')

    # cache tree numbers
    tree_number_to_descriptor_ui = {
        tree_number: descriptor['descriptor_ui']
        for descriptor in rv
        for tree_number in descriptor['tree_numbers']
    }
    logger.debug(f'got {len(tree_number_to_descriptor_ui)} tree mappings')

    # add in parents to each descriptor based on their tree numbers
    for descriptor in rv:
        parents_descriptor_uis = set()
        for tree_number in descriptor['tree_numbers']:
            try:
                parent_tn, self_tn = tree_number.rsplit('.', 1)
            except ValueError:
                logger.debug('No dot for %s', tree_number)
                continue

            parent_descriptor_ui = tree_number_to_descriptor_ui.get(parent_tn)
            if parent_descriptor_ui is not None:
                parents_descriptor_uis.add(parent_descriptor_ui)
            else:
                logger.debug('missing tree number: %s', parent_tn)

        descriptor['parents'] = list(parents_descriptor_uis)

    return rv


def get_scope_note(term) -> Optional[str]:
    """Get the scope note from the preferred concept in a term's record."""
    for concept in term['concepts']:
        if 'ScopeNote' in concept:
            return concept['ScopeNote']


def get_descriptor_record(element: Element) -> Dict[str, Any]:
    """Get descriptor records from the main element."""
    return {
        'descriptor_ui': element.findtext('DescriptorUI'),
        'name': element.findtext('DescriptorName/String'),
        'tree_numbers': sorted({
            x.text
            for x in element.findall('TreeNumberList/TreeNumber')
        }),
        'concepts': get_concept_records(element),
        # TODO handle AllowableQualifiersList
        # TODO add ScopeNote as description
    }


def get_concept_records(element: Element) -> List[Mapping[str, Any]]:
    """Get concepts from a record."""
    return [
        get_concept_record(concept)
        for concept in element.findall('ConceptList/Concept')
    ]


def get_concept_record(concept):
    """Get a single MeSH concept record."""
    return {
        'concept_ui': concept.findtext('ConceptUI'),
        'name': concept.findtext('ConceptName/String'),
        'semantic_types': list({
            x.text
            for x in concept.findall('SemanticTypeList/SemanticType/SemanticTypeUI')
        }),
        'ScopeNote': concept.findtext('ScopeNote'),
        'terms': get_term_records(concept),
        # TODO handle ConceptRelationList
        **concept.attrib,
    }


def get_term_records(element: Element) -> List[Mapping[str, Any]]:
    """Get all of the terms for a concept."""
    return [
        get_term_record(term)
        for term in element.findall('TermList/Term')
    ]


def get_term_record(term):
    """Get a single MeSH term record."""
    return {
        'term_ui': term.findtext('TermUI'),
        'name': term.findtext('String'),
        **term.attrib,
    }


def _get_descriptor_qualifiers(descriptor: Element) -> List[Mapping[str, str]]:
    return [
        {
            'qualifier_ui': qualifier.findtext('QualifierUI'),
            'name': qualifier.findtext('QualifierName/String'),
        }
        for qualifier in descriptor.findall('AllowableQualifiersList/AllowableQualifier/QualifierReferredTo')
    ]


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Extraction of mappings from OBO documents."""

from .extract_names import get_id_name_mapping, get_name_id_mapping  # noqa: F401
from .extract_synonyms import get_synonyms  # noqa: F401
from .extract_xrefs import get_all_xrefs, get_xrefs, iterate_xrefs_from_graph  # noqa: F401



# -*- coding: utf-8 -*-

"""Converter for HGNC Gene Families."""

from collections import defaultdict
from typing import Iterable, List, Mapping

import pandas as pd
from tqdm import tqdm

from ..path_utils import ensure_path
from ..struct import Obo, Reference, Synonym, SynonymTypeDef, Term

PREFIX = 'hgnc.genefamily'
FAMILIES_URL = 'ftp://ftp.ebi.ac.uk/pub/databases/genenames/new/csv/genefamily_db_tables/family.csv'
HIERARCHY_URL = 'ftp://ftp.ebi.ac.uk/pub/databases/genenames/new/csv/genefamily_db_tables/hierarchy.csv'

symbol_type = SynonymTypeDef(id='symbol', name='symbol')


def get_obo() -> Obo:
    """Get HGNC Gene Families as OBO."""
    terms = list(get_terms())
    hierarchy = get_hierarchy()

    id_to_term = {term.reference.identifier: term for term in terms}
    for child_id, parent_ids in hierarchy.items():
        child = id_to_term[child_id]
        for parent_id in parent_ids:
            parent: Term = id_to_term[parent_id]
            child.parents.append(Reference(
                prefix=PREFIX,
                identifier=parent_id,
                name=parent.name,
            ))

    return Obo(
        ontology=PREFIX,
        name='HGNC Gene Families',
        terms=terms,
        synonym_typedefs=[symbol_type],
        auto_generated_by='bio2obo:hgnc.genefamily',
    )


def get_hierarchy() -> Mapping[str, List[str]]:
    """Get the HGNC Gene Families hierarchy as a dictionary."""
    path = ensure_path(PREFIX, HIERARCHY_URL)
    df = pd.read_csv(path, dtype={'parent_fam_id': str, 'child_fam_id': str})
    d = defaultdict(list)
    for parent_id, child_id in df.values:
        d[child_id].append(parent_id)
    return dict(d)


COLUMNS = ['id', 'abbreviation', 'name', 'pubmed_ids', 'desc_comment', 'desc_go']


def get_terms() -> Iterable[Term]:
    """Get the HGNC Gene Family terms."""
    path = ensure_path(PREFIX, FAMILIES_URL)
    df = pd.read_csv(path, dtype={'id': str})

    it = tqdm(df[COLUMNS].values, desc=f'Mapping {PREFIX}')
    for hgncgenefamily_id, symbol, name, pubmed_ids, definition, desc_go in it:
        if pubmed_ids and pd.notna(pubmed_ids):
            provenance = [Reference(prefix='pubmed', identifier=s.strip()) for s in pubmed_ids.split(',')]
        else:
            provenance = []

        if not definition or pd.isna(definition):
            definition = ''

        xrefs = []
        if desc_go and pd.notna(desc_go):
            go_id = desc_go[len('http://purl.uniprot.org/go/'):]
            xrefs.append(Reference(prefix='go', identifier=go_id))

        synonyms = []
        if symbol and pd.notna(symbol):
            synonyms.append(Synonym(name=symbol, type=symbol_type))

        term = Term(
            name=name,
            reference=Reference(prefix=PREFIX, identifier=hgncgenefamily_id),
            definition=definition,
            provenance=provenance,
            xrefs=xrefs,
            synonyms=synonyms,
        )
        yield term


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""A python package for handling and generating OBO."""

from .getters import get_obo_graph, get_obo_graph_by_prefix, get_obo_graph_by_url  # noqa: F401
from .mappings.extract_names import get_id_name_mapping, get_name_id_mapping  # noqa: F401
from .mappings.extract_synonyms import get_synonyms  # noqa: F401
from .mappings.extract_xrefs import get_all_xrefs, get_xrefs, iterate_xrefs_from_graph  # noqa: F401
from .path_utils import ensure_path  # noqa: F401
from .sources import CONVERTED, get_converted_obo, get_converted_obos  # noqa: F401
from .struct import Obo, Reference, Synonym, SynonymTypeDef, Term, TypeDef, get_terms_from_graph  # noqa: F401
from .version import get_version  # noqa: F401



# -*- coding: utf-8 -*-

"""I/O utilities."""

import gzip
import logging
import time
from collections import defaultdict
from typing import Iterable, List, Mapping, Tuple, TypeVar
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

__all__ = [
    'split_tab_pair',
    'open_map_tsv',
    'open_multimap_tsv',
    'multidict',
    'write_map_tsv',
    'write_multimap_tsv',
    'write_iterable_tsv',
    'parse_xml_gz',
]

logger = logging.getLogger(__name__)

X = TypeVar('X')
Y = TypeVar('Y')


def split_tab_pair(x: str) -> Tuple[str, str]:
    """Split a pair of elements by a tab."""
    a, b = x.strip().split('\t')
    return a, b


def open_map_tsv(path: str) -> Mapping[str, str]:
    """Load a mapping TSV file into a dictionary."""
    with open(path) as file:
        next(file)  # throw away header
        return dict(split_tab_pair(line) for line in file)


def open_multimap_tsv(path) -> Mapping[str, List[str]]:
    """Load a mapping TSV file that has multiple mappings for each."""
    rv = defaultdict(list)
    with open(path) as file:
        next(file)  # throw away header
        for line in file:
            key, value = split_tab_pair(line)
            rv[key].append(value)
    return dict(rv)


def multidict(pairs: Iterable[Tuple[X, Y]]) -> Mapping[X, List[Y]]:
    """Accumulate a multidict from a list of pairs."""
    rv = defaultdict(list)
    for key, value in pairs:
        rv[key].append(value)
    return dict(rv)


def write_map_tsv(*, path: str, header: Iterable[str], rv: Mapping[str, str], sep: str = '\t') -> None:
    """Write a mapping dictionary to a TSV file."""
    write_iterable_tsv(path=path, header=header, it=rv.items(), sep=sep)


def write_multimap_tsv(*, path: str, header: Iterable[str], rv: Mapping[str, List[str]], sep: str = '\t') -> None:
    """Write a multiple mapping dictionary to a TSV file."""
    it = (
        (key, value)
        for key, values in rv.items()
        for value in values
    )
    write_iterable_tsv(path=path, header=header, it=it, sep=sep)


def write_iterable_tsv(*, path: str, header: Iterable[str], it: Iterable[Tuple[str, str]], sep: str = '\t') -> None:
    """Write a mapping dictionary to a TSV file."""
    with open(path, 'w') as file:
        print(*header, sep=sep, file=file)
        for key, value in sorted(it):
            print(key, value, sep=sep, file=file)


def parse_xml_gz(path: str) -> Element:
    """Parse an XML file from a path to a GZIP file."""
    t = time.time()
    logger.info('parsing xml from %s', path)
    with gzip.open(path) as file:
        tree = ElementTree.parse(file)
    logger.info('parsed xml in %.2f seconds', time.time() - t)
    return tree.getroot()



# -*- coding: utf-8 -*-

"""Converter for HGNC."""

import json
import logging
from typing import Iterable

from tqdm import tqdm

from ..path_utils import ensure_path
from ..struct import Obo, Reference, Synonym, SynonymTypeDef, Term, from_species

logger = logging.getLogger(__name__)

PREFIX = 'hgnc'
DEFINITIONS_URL = 'ftp://ftp.ebi.ac.uk/pub/databases/genenames/new/json/hgnc_complete_set.json'

previous_symbol_type = SynonymTypeDef(id='previous_symbol', name='previous symbol')
alias_symbol_type = SynonymTypeDef(id='alias_symbol', name='alias symbol')
previous_name_type = SynonymTypeDef(id='previous_name', name='previous name')
alias_name_type = SynonymTypeDef(id='alias_name', name='alias name')

gene_xrefs = [
    ('ensembl', 'ensembl_gene_id'),
    ('ncbigene', 'entrez_id'),
    ('cosmic', 'cosmic'),
    ('vega', 'vega_id'),
    ('ucsc', 'ucsc_id'),
    ('merops', 'merops'),
    ('lncipedia', 'lncipedia'),
    ('orphanet', 'orphanet'),
    ('pseudogene', 'pseudogene.org'),
    ('ena', 'ena'),
    ('refseq', 'refseq_accession'),
    ('mgi', 'mgd_id'),
    ('ccds', 'ccds_id'),
    ('rgd', 'rgd_id'),
    ('omim', 'omim_id'),
    ('uniprot', 'uniprot_ids'),
    ('ec-code', 'enzyme_id'),
    ('rnacentral', 'rna_central_id'),
    ('mirbase', 'mirbase'),
    ('snornabase', 'snornabase'),
]


def get_obo() -> Obo:
    """Get HGNC as OBO."""
    terms = list(get_terms())
    return Obo(
        ontology=PREFIX,
        name='HGNC',
        terms=terms,
        typedefs=[from_species],
        synonym_typedefs=[previous_name_type, previous_symbol_type, alias_name_type, alias_symbol_type],
        auto_generated_by='bio2obo:hgnc',
    )


def get_terms() -> Iterable[Term]:
    """Get HGNC terms."""
    unhandled = set()
    path = ensure_path(PREFIX, DEFINITIONS_URL)
    with open(path) as file:
        entries = json.load(file)['response']['docs']

    for entry in tqdm(entries, desc=f'Mapping {PREFIX}'):
        name, symbol, identifier = entry.pop('name'), entry.pop('symbol'), entry.pop('hgnc_id')[len('HGNC:'):]

        xrefs = []
        for xref_prefix, key in gene_xrefs:
            xref_identifiers = entry.pop(key, None)
            if xref_identifiers is None:
                continue
            if not isinstance(xref_identifiers, list):
                xref_identifiers = [xref_identifiers]
            for xref_identifier in xref_identifiers:
                xrefs.append(Reference(prefix=xref_prefix, identifier=str(xref_identifier)))

        provenance = [
            Reference(prefix='pubmed', identifier=str(pubmed_id))
            for pubmed_id in entry.pop('pubmed_id', [])
        ]

        gene_group_ids = entry.pop('gene_group_id', [])
        gene_groups = entry.pop('gene_group', [])
        parents = [
            Reference(
                prefix='hgnc.genefamily',
                identifier=str(hgncgenefamily_id),
                name=gene_group_label,
            )
            for hgncgenefamily_id, gene_group_label in zip(gene_group_ids, gene_groups)
        ]

        synonyms = []
        for alias_symbol in entry.pop('alias_symbol', []):
            synonyms.append(Synonym(name=alias_symbol, type=alias_symbol_type))
        for alias_name in entry.pop('alias_name', []):
            synonyms.append(Synonym(name=alias_name, type=alias_name_type))
        for previous_symbol in entry.pop('previous_symbol', []):
            synonyms.append(Synonym(name=previous_symbol, type=previous_symbol_type))
        for previous_name in entry.pop('prev_name', []):
            synonyms.append(Synonym(name=previous_name, type=previous_name_type))

        term = Term(
            name=symbol,
            definition=name,
            reference=Reference(prefix=PREFIX, identifier=identifier),
            xrefs=xrefs,
            provenance=provenance,
            parents=parents,
            synonyms=synonyms,
        )
        term.append_relationship(from_species, Reference(prefix='taxonomy', identifier='9606', name='Homo sapiens'))

        unhandled.update(set(entry))
        yield term

    # logger.warning('Unhandled:')
    # for u in sorted(unhandled):
    #     logger.warning(u)


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Extract registry information."""

from .registries import (
    get_curated_registry, get_metaregistry, get_miriam, get_namespace_synonyms, get_obofoundry, get_ols,
)

__all__ = [
    'get_curated_registry',
    'get_metaregistry',
    'get_miriam',
    'get_namespace_synonyms',
    'get_obofoundry',
    'get_ols',
]



# -*- coding: utf-8 -*-

"""Getters for OBO content."""

from typing import Iterable, Mapping, Optional, Tuple

import networkx as nx

from ..cache_utils import cached_mapping
from ..getters import get_obo_graph
from ..path_utils import prefix_directory_join

__all__ = [
    'get_id_name_mapping',
    'get_name_id_mapping',
]


def get_id_name_mapping(prefix: str, url: Optional[str] = None) -> Mapping[str, str]:
    """Get an identifier to name mapping for the OBO file."""
    path = prefix_directory_join(prefix, f"{prefix}.mapping.tsv")

    @cached_mapping(path=path, header=[f'{prefix}_id', 'name'])
    def _get_id_name_mapping() -> Mapping[str, str]:
        graph = get_obo_graph(prefix, url=url)
        return dict(_iterate_identifier_names(graph, prefix))

    return _get_id_name_mapping()


def get_name_id_mapping(prefix: str, url: Optional[str] = None) -> Mapping[str, str]:
    """Get a name to identifier mapping for the OBO file."""
    d = get_id_name_mapping(prefix=prefix, url=url)
    return {v: k for k, v in d.items()}


def _iterate_identifier_names(graph: nx.MultiDiGraph, prefix: str) -> Iterable[Tuple[str, str]]:
    for node, data in graph.nodes(data=True):
        identifier = node[len(f'{prefix}:'):]
        name = data.get('name')
        if name is None:
            continue
        yield identifier, name



# -*- coding: utf-8 -*-

"""Utilities for extracting xrefs."""

import logging
import os
from collections import defaultdict
from typing import Iterable, List, Mapping, Optional, Tuple

import networkx as nx
import pandas as pd
from tqdm import tqdm

from ..cache_utils import cached_df, cached_multidict
from ..getters import get_obo_graph
from ..io_utils import multidict
from ..path_utils import prefix_directory_join
from ..registries import get_curated_registry, get_namespace_synonyms

__all__ = [
    'iterate_xrefs_from_graph',
    'get_xrefs',
    'get_all_xrefs',
]

logger = logging.getLogger(__name__)

HERE = os.path.abspath(os.path.dirname(__file__))

METAREGISTRY = get_curated_registry()

#: Xrefs starting with these prefixes will be ignored
XREF_PREFIX_BLACKLIST = set(METAREGISTRY['blacklists']['prefix'])
XREF_SUFFIX_BLACKLIST = set(METAREGISTRY['blacklists']['suffix'])
#: Xrefs matching these will be ignored
XREF_BLACKLIST = set(METAREGISTRY['blacklists']['full'])

PREFIX_REMAP = METAREGISTRY['remappings']['prefix']

ALLOWED_UNNORM = set(METAREGISTRY['database'])

SYNONYM_TO_KEY = get_namespace_synonyms()
UNHANDLED_NAMESPACES = defaultdict(list)
UBERON_UNHANDLED = defaultdict(list)


def get_all_xrefs(prefix: str, url: Optional[str] = None) -> pd.DataFrame:
    """Get all xrefs."""
    path = prefix_directory_join(prefix, f"{prefix}_mappings.tsv")
    dtype = {
        'source_ns': str, 'source_id': str, 'xref_ns': str, 'xref_id': str,
    }

    @cached_df(path=path, dtype=dtype)
    def _df_getter() -> pd.DataFrame:
        graph = get_obo_graph(prefix, url=url)
        logger.info('writing %s mapping to %s', prefix, path)
        return pd.DataFrame(
            list(iterate_xrefs_from_graph(graph)),
            columns=['source_ns', 'source_id', 'target_ns', 'target_id'],
        )

    return _df_getter()


def get_xrefs(prefix: str, xref_prefix: str, url: Optional[str] = None) -> Mapping[str, List[str]]:
    """Get xrefs to a given target."""
    path = prefix_directory_join(prefix, f"{prefix}_{xref_prefix}_mappings.tsv")
    header = [f'{prefix}_id', f'{xref_prefix}_id']

    @cached_multidict(path=path, header=header)
    def _get_multidict() -> Mapping[str, List[str]]:
        graph = get_obo_graph(prefix, url=url)
        return multidict(
            (head_id, xref_id)
            for head_ns, head_id, xref_ns, xref_id in iterate_xrefs_from_graph(graph)
            if head_ns == prefix and xref_ns == xref_prefix
        )

    return _get_multidict()


def iterate_xrefs_from_graph(graph: nx.Graph, use_tqdm: bool = True) -> Iterable[Tuple[str, str, str, str]]:
    """Iterate over cross references in the graph."""
    it = graph.nodes(data=True)
    if use_tqdm:
        it = tqdm(it, desc=f'Extracting xrefs from {graph}')
    for node, data in it:
        node = node.strip()

        if node in XREF_BLACKLIST:
            continue

        # Skip node if it has a blacklisted prefix
        for prefix in XREF_PREFIX_BLACKLIST:
            if node.startswith(prefix):
                continue

        # Skip node if it has a blacklisted suffix
        for suffix in XREF_SUFFIX_BLACKLIST:
            if node.endswith(suffix):
                continue

        # Remap node's prefix (if necessary)
        for prefix, new_prefix in PREFIX_REMAP.items():
            if node.startswith(prefix):
                node = new_prefix + node[len(prefix):]

        try:
            head_ns, head_id = node.split(':', 1)
        except ValueError:  # skip nodes that don't look like normal CURIEs
            # logger.info(f'skipping: {node}')
            continue

        norm_head_ns = normalize_namespace(head_ns, node, None)
        if not norm_head_ns:
            continue

        # TODO check if synonyms are also written like CURIEs,
        # ... not that they should be

        for xref in data.get('xref', []):
            xref = xref.strip()

            if (
                node == xref
                or any(xref.startswith(x) for x in XREF_PREFIX_BLACKLIST)
                or xref in XREF_BLACKLIST
                or ':' not in xref
            ):
                continue  # sometimes xref to self... weird

            for prefix, new_prefix in PREFIX_REMAP.items():
                if xref.startswith(prefix):
                    xref = new_prefix + xref[len(prefix):]

            split_space = ' ' in xref
            if split_space:
                _xref_split = xref.split(' ', 1)
                if _xref_split[1][0] not in {'"', '('}:
                    logger.warning(f'Problem with space in xref {node} {xref}')
                    continue
                xref = _xref_split[0]

            try:
                xref_ns, xref_id = xref.split(':', 1)
            except ValueError:
                if split_space:
                    logger.warning(f'problem splitting after space split {node} {xref}')
                else:
                    logger.warning(f'problem splitting {node} {xref}')

                continue

            norm_xref_ns = normalize_namespace(xref_ns, node, xref)
            if not norm_xref_ns:
                continue

            yield norm_head_ns, head_id, norm_xref_ns, xref_id


def normalize_namespace(namespace: str, curie: str, xref=None) -> Optional[str]:
    """Normalize a namespace and return, if possible."""
    for string_transformation in (lambda x: x, str.lower, str.upper, str.casefold):
        namespace_transformed = string_transformation(namespace)
        if namespace_transformed in SYNONYM_TO_KEY:
            return SYNONYM_TO_KEY[namespace_transformed]

    if curie.startswith('UBERON:'):  # uberon has tons of xrefs to anatomical features. skip them
        UBERON_UNHANDLED[namespace].append((curie, xref))
    else:
        UNHANDLED_NAMESPACES[namespace].append((curie, xref))



# -*- coding: utf-8 -*-

"""Utilities for OBO files."""

import logging
import os
from typing import Optional
from urllib.request import urlretrieve

import networkx as nx

from .cache_utils import ensure_obo_graph
from .constants import CURATED_URLS
from .path_utils import ensure_path, get_prefix_obo_path
from .registries import get_obofoundry
from .sources import CONVERTED, get_converted_obo

__all__ = [
    'get_obo_graph',
    'get_obo_graph_by_url',
    'get_obo_graph_by_prefix',
]

logger = logging.getLogger(__name__)


class MissingOboBuild(RuntimeError):
    """Raised when OBOFoundry doesn't track an OBO file, but only has OWL."""


def get_obo_graph(prefix: str, *, url: Optional[str] = None) -> nx.MultiDiGraph:
    """Get the OBO file by prefix or URL."""
    if prefix in CONVERTED:
        get_converted_obo(prefix).write_default()
    if url is None:
        return get_obo_graph_by_prefix(prefix)
    else:
        return get_obo_graph_by_url(prefix, url)


def get_obo_graph_by_url(prefix: str, url: str) -> nx.MultiDiGraph:
    """Get the OBO file as a graph using the given URL and cache if not already."""
    path = get_prefix_obo_path(prefix)
    if not os.path.exists(path):
        logger.info('downloading %s OBO from %s', prefix, url)
        urlretrieve(url, path)
    return ensure_obo_graph(path)


def get_obo_graph_by_prefix(prefix: str) -> nx.MultiDiGraph:
    """Get the OBO file as a graph using the OBOFoundry registry URL and cache if not already."""
    path = ensure_obo_path(prefix)
    return ensure_obo_graph(path=path)


def ensure_obo_path(prefix: str) -> str:
    """Get the path to the OBO file and download if missing."""
    if prefix in CURATED_URLS:
        return ensure_path(prefix, CURATED_URLS[prefix])

    path = get_prefix_obo_path(prefix)
    if os.path.exists(path):
        return path

    obofoundry = get_obofoundry(mappify=True)
    entry = obofoundry.get(prefix)
    if entry is None:
        raise ValueError(f'OBO Foundry is missing the prefix: {prefix}')

    build = entry.get('build')
    if build is None:
        raise MissingOboBuild(f'OBO Foundry is missing a build for: {prefix}')
    url = build['source_url']
    return ensure_path(prefix, url)



# -*- coding: utf-8 -*-

"""Parser functions for OBO documents."""

from typing import Any, List, Mapping

import networkx as nx

from .struct import Reference, Term

__all__ = [
    'get_terms_from_graph',
]


def get_terms_from_graph(graph: nx.Graph) -> List[Term]:
    """Get all of the terms from a OBO graph."""
    ontology = graph.graph['ontology']

    #: Identifiers to references
    references = {
        node: Reference(prefix=ontology, identifier=node, name=data['name'])
        for node, data in graph.nodes(data=True)
    }

    def _make_term(node: str, data: Mapping[str, Any]) -> Term:
        reference = references[node]
        return Term(
            reference=reference,
            name=reference.name,
            definition=data['def'],
            parents=list(_get_parents(data)),
        )

    def _get_parents(data: Mapping[str, Any]):
        for parent in data.get('is_a', []):
            # May have to add more logic here later
            yield references[parent]

    terms = []
    for node, data in graph.nodes(data=True):
        term = _make_term(node, data)
        terms.append(term)

    return terms



# -*- coding: utf-8 -*-

"""Convert ExPASy to OBO."""

import logging
from collections import defaultdict
from typing import Dict, Iterable, Mapping

from ..path_utils import ensure_path
from ..struct import Obo, Reference, Synonym, Term

PREFIX = 'ec-code'
EXPASY_DATABASE_URL = 'ftp://ftp.expasy.org/databases/enzyme/enzyme.dat'
EXPASY_TREE_URL = 'ftp://ftp.expasy.org/databases/enzyme/enzclass.txt'

logger = logging.getLogger(__name__)

#: The identifier of the entry (One)
ID = 'ID'
#: Description (One)
DE = 'DE'
#: Additional names/synonyms (Many)
AN = 'AN'
#: Chemical Reaction String (One)
CA = 'CA'
#: Comments (One - consider as free text)
CC = 'CC'
#: List of cofactors? (Many)
CF = 'CF'
#: ProSite Identifier (optional) (Many)
PR = 'PR'
#: Reference to UniProt or SwissProt (Many)
DR = 'DR'


def get_obo() -> Obo:
    """Get ExPASy as OBO."""
    terms = list(get_terms())
    return Obo(
        ontology=PREFIX,
        name='ExPASy Enzyme Nomenclature',
        terms=terms,
    )


def get_terms() -> Iterable[Term]:
    """Get the ExPASy terms."""
    tree_path = ensure_path(PREFIX, EXPASY_TREE_URL)
    with open(tree_path) as file:
        tree = get_tree(file)

    terms: Dict[str, Term] = {}
    child_to_parents = defaultdict(list)
    for ec_code, data in tree.items():
        terms[ec_code] = Term(
            name=data['name'],
            reference=Reference(prefix=PREFIX, identifier=ec_code),
        )
        for child_data in data.get('children', []):
            child_ec_code = child_data['identifier']
            child_to_parents[child_ec_code].append(ec_code)

    for child_ec_code, parents_ec_codes in child_to_parents.items():
        terms[child_ec_code].parents = [
            terms[parent_ec_code].reference
            for parent_ec_code in parents_ec_codes
        ]

    database_path = ensure_path(PREFIX, EXPASY_DATABASE_URL)
    with open(database_path) as file:
        data = get_database(file)

    ec_code_to_alt_ids = {}
    for ec_code, data in data.items():
        parent_ec_code = data['parent']['identifier']
        parent_term = terms[parent_ec_code]

        synonyms = [
            Synonym(name=synonym)
            for synonym in data.get('synonyms', [])
        ]
        if data['alt_ids']:
            alt_ids = data['alt_ids'][0].rstrip('.')
            if 'and' not in alt_ids:
                ec_code_to_alt_ids[ec_code] = [
                    alt_ids,
                ]
            else:
                ec_code_to_alt_ids[ec_code] = [
                    alt_id.rstrip(',')
                    for alt_id in alt_ids.split(' ')
                    if alt_id != 'and'
                ]

        xrefs = []
        for domain in data.get('domains', []):
            xrefs.append(Reference(prefix=domain['namespace'], identifier=domain['identifier']))
        for protein in data.get('proteins', []):
            xrefs.append(Reference(prefix=protein['namespace'], identifier=protein['identifier'], name=protein['name']))

        concept = data['concept']
        try:
            name = concept['name']
        except KeyError:
            continue
            # raise

        terms[ec_code] = Term(
            name=name,
            reference=Reference(prefix=PREFIX, identifier=ec_code),
            parents=[parent_term.reference],
            synonyms=synonyms,
            xrefs=xrefs,
        )

    return terms.values()


"""TREE"""


def normalize_expasy_id(expasy_id: str) -> str:
    """Return a standardized ExPASy identifier string.

    :param expasy_id: A possibly non-normalized ExPASy identifier
    """
    return expasy_id.replace(" ", "")


def give_edge(head_str: str):
    """Return a (parent, child) tuple for given id."""
    head_str = normalize_expasy_id(head_str)
    nums = head_str.split('.')
    for i, obj in enumerate(nums):
        nums[i] = obj.strip()

    while '-' in nums:
        nums.remove('-')

    level = len(nums)

    if level == 1:
        return level, None, "{}.-.-.-".format(nums[0])

    if level == 2:
        return (
            level,
            normalize_expasy_id("{}. -. -.-".format(nums[0])),
            normalize_expasy_id("{}.{:>2}. -.-".format(nums[0], nums[1])),
        )

    if level == 3:
        return (
            level,
            normalize_expasy_id("{}.{:>2}. -.-".format(nums[0], nums[1])),
            normalize_expasy_id("{}.{:>2}.{:>2}.-".format(nums[0], nums[1], nums[2])),
        )

    if level == 4:
        return (
            level,
            normalize_expasy_id("{}.{:>2}.{:>2}.-".format(nums[0], nums[1], nums[2])),
            normalize_expasy_id("{}.{:>2}.{:>2}.{}".format(nums[0], nums[1], nums[2], nums[3])),
        )


def _process_line(line, graph):
    line.rstrip('\n')
    if not line[0].isnumeric():
        return
    head = line[:10]
    l_nums, parent, child = give_edge(head)
    name = line[11:]
    name = name.strip().strip('.')
    graph.add_node(child, description=name)
    if parent is not None:
        graph.add_edge(parent, child)


def get_tree(lines: Iterable[str]):
    """Get the ExPASy tree mapping."""
    rv = {}
    for line in lines:
        if not line[0].isnumeric():
            continue
        head = line[:10]
        level, parent_expasy_id, expasy_id = give_edge(head)
        name = line[11:]
        name = name.strip().strip('.')

        rv[expasy_id] = {
            'concept': {
                'namespace': 'ec-code',
                'identifier': expasy_id,
            },
            'name': name,
            'level': level,
            'children': [],
        }
        if parent_expasy_id is not None:
            rv[expasy_id]['parent'] = {
                'namespace': 'ec-code',
                'identifier': parent_expasy_id,
            }
            rv[parent_expasy_id]['children'].append(rv[expasy_id]['concept'])

    return rv


"""DATABASE"""


def get_database(lines: Iterable[str]) -> Mapping:
    """Parse the ExPASy database file and returns a list of enzyme entry dictionaries.

    :param lines: An iterator over the ExPASy database file or file-like
    """
    rv = {}
    for groups in group_by_id(lines):
        _, expasy_id = groups[0]

        rv[expasy_id] = ec_data_entry = {
            'concept': {
                'namespace': 'ec-code',
                'identifier': expasy_id,
            },
            'parent': {
                'namespace': 'ec-code',
                'identifier': expasy_id.rsplit('.', 1)[0] + '.-',
            },
            'synonyms': [],
            'cofactors': [],
            'domains': [],
            'proteins': [],
            'alt_ids': [],
        }

        for descriptor, value in groups[1:]:
            if descriptor == '//':
                continue
            elif descriptor == DE and value == 'Deleted entry.':
                continue
            elif descriptor == DE and value.startswith('Transferred entry: '):
                value = value[len('Transferred entry: '):].rstrip()
                ec_data_entry['transfer_id'] = value
            elif descriptor == DE:
                ec_data_entry['concept']['name'] = value.rstrip('.')
            elif descriptor == AN:
                ec_data_entry['synonyms'].append(value.rstrip('.'))
            elif descriptor == PR:
                value = value[len('PROSITE; '):-1]  # remove trailing comma
                ec_data_entry['domains'].append({
                    'namespace': 'prosite',
                    'identifier': value,
                })
            elif descriptor == DR:
                for uniprot_entry in value.replace(' ', '').split(';'):
                    if not uniprot_entry:
                        continue
                    uniprot_id, uniprot_accession = uniprot_entry.split(',')
                    ec_data_entry['proteins'].append(dict(
                        namespace='uniprot',
                        name=uniprot_accession,
                        identifier=uniprot_id,
                    ))

    for expasy_id, data in rv.items():
        transfer_id = data.pop('transfer_id', None)
        if transfer_id is not None:
            rv[expasy_id]['alt_ids'].append(transfer_id)

    return rv


def group_by_id(lines):
    """Group lines by identifier."""
    groups = []
    for line in lines:  # TODO replace with itertools.groupby
        line = line.strip()

        if line.startswith('ID'):
            groups.append([])

        if not groups:
            continue

        descriptor = line[:2]
        value = line[5:]

        groups[-1].append((descriptor, value))

    return groups


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Converter for miRBase."""

import gzip
import logging
from typing import Iterable, List, Mapping

from tqdm import tqdm

from ..cache_utils import cached_mapping
from ..path_utils import ensure_df, ensure_path, prefix_directory_join
from ..struct import Obo, Reference, Synonym, Term, TypeDef, from_species

logger = logging.getLogger(__name__)

PREFIX = 'mirbase'
MIRBASE_MATURE_PREFIX = 'mirbase.mature'
VERSION = '22.1'
DEFINITIONS_URL = f'ftp://mirbase.org/pub/mirbase/{VERSION}/miRNA.dat.gz'
ALIASES_URL = f'ftp://mirbase.org/pub/mirbase/{VERSION}/aliases.txt.gz'
SPECIES_URL = f'ftp://mirbase.org/pub/mirbase/{VERSION}/organisms.txt.gz'

has_mature = TypeDef(id='has_mature', name='has mature miRNA')

xref_mapping = {
    'entrezgene': 'ncbigene',
}


def get_obo() -> Obo:
    """Get miRBase as OBO."""
    terms = get_terms()
    return Obo(
        ontology=PREFIX,
        name='miRBase',
        terms=terms,
        typedefs=[from_species, has_mature],
        data_version=VERSION,
        auto_generated_by=f'bio2obo:{PREFIX}',
    )


def get_terms() -> List[Term]:
    """Parse miRNA data from filepath and convert it to dictionary."""
    definitions_path = ensure_path(PREFIX, DEFINITIONS_URL)

    file_handle = (
        gzip.open(definitions_path, 'rt')
        if definitions_path.endswith('.gz') else
        open(definitions_path)
    )
    with file_handle as file:
        return list(_process_definitions_lines(file))


def _prepare_organisms():
    df = ensure_df(PREFIX, SPECIES_URL, sep='\t', dtype={'#NCBI-taxid': str})
    return {
        division: (taxonomy_id, name)
        for _, division, name, _tree, taxonomy_id in df.values
    }


def _prepare_aliases() -> Mapping[str, List[str]]:
    df = ensure_df(PREFIX, ALIASES_URL, sep='\t')
    return {
        mirbase_id: [s.strip() for s in synonyms.split(';') if s and s.strip()]
        for mirbase_id, synonyms in df.values
    }


def _process_definitions_lines(lines: Iterable[str]) -> Iterable[Term]:
    """Process the lines of the definitions file."""
    organisms = _prepare_organisms()
    aliases = _prepare_aliases()

    groups = []

    for line in lines:  # TODO replace with itertools.groupby
        if line.startswith('ID'):
            listnew = []
            groups.append(listnew)
        groups[-1].append(line)

    for group in tqdm(groups, desc='Parsing miRBase'):
        name = group[0][5:23].strip()
        qualifier, dtype, species_code, length = map(str.strip, group[0][23:].strip().rstrip('.').split(';'))
        identifier = group[2][3:-2].strip()
        definition = group[4][3:-1].strip()

        synonyms = [
            Synonym(name=alias)
            for alias in aliases.get(identifier, [])
            if alias != name
        ]

        species_identifier, species_name = organisms[species_code]
        species = Reference(
            prefix='taxonomy',
            identifier=species_identifier,
            name=species_name,
        )

        mature_mirna_lines = [
            i
            for i, element in enumerate(group)
            if 'FT   miRNA    ' in element
        ]

        matures = []
        for index in mature_mirna_lines:
            # location = group[index][10:-1].strip()
            accession = group[index + 1][33:-2]
            product = group[index + 2][31:-2]
            product_reference = Reference(
                prefix=MIRBASE_MATURE_PREFIX,
                identifier=accession,
                name=product,
            )
            if product.endswith('3p') or product.endswith('5p'):
                matures.append(product_reference)
            else:
                pass
                # logger.warning(f'Whats going on {group[index]}')

        xrefs = []
        for line in group:
            if not line.startswith('DR'):
                continue
            line = line[len('DR   '):].strip().rstrip('.')
            xref_prefix, xref_identifier, xref_label = map(str.strip, line.split(';'))
            xref_prefix = xref_prefix.lower()
            xref_prefix = xref_mapping.get(xref_prefix, xref_prefix)
            xrefs.append(Reference(prefix=xref_prefix, identifier=xref_identifier,
                                   name=xref_label or None))

        # TODO add pubmed references

        term = Term(
            name=name,
            reference=Reference(prefix=PREFIX, identifier=identifier),
            definition=definition,
            xrefs=xrefs,
            synonyms=synonyms,
        )
        term.append_relationship(from_species, species)
        term.extend_relationship(has_mature, matures)

        yield term


@cached_mapping(
    path=prefix_directory_join(PREFIX, f'{PREFIX}.mature_to_{PREFIX}.tsv'),
    header=['mirbase.mature_id', 'mirbase_id'],
)
def get_mature_to_premature() -> Mapping[str, str]:
    """Get a mapping from mature miRNAs to their parents."""
    return {
        mature.identifier: term.identifier
        for term in get_terms()
        for mature in term.get_relationships(has_mature)
    }


@cached_mapping(
    path=prefix_directory_join(PREFIX, f'{PREFIX}.mature_mapping.tsv'),
    header=['mirbase.mature_id', 'name'],
)
def get_mature_id_to_name() -> Mapping[str, str]:
    """Get a mapping from mature miRNAs to their parents."""
    return {
        mature.identifier: mature.name
        for term in get_terms()
        for mature in term.get_relationships(has_mature)
    }


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Version information for PyOBO."""

__all__ = [
    'VERSION',
    'get_version',
]

VERSION = '0.0.5'


def get_version() -> str:
    """Get the software version of PyOBO."""
    return VERSION



# -*- coding: utf-8 -*-

"""Utilities for extracting synonyms."""

import logging
from typing import Iterable, List, Mapping, Optional, Tuple

import networkx as nx

from ..cache_utils import cached_multidict
from ..getters import get_obo_graph
from ..io_utils import multidict
from ..path_utils import prefix_directory_join

__all__ = [
    'get_synonyms',
]

logger = logging.getLogger(__name__)


def get_synonyms(prefix: str, url: Optional[str] = None) -> Mapping[str, List[str]]:
    """Get the OBO file and output a synonym dictionary."""
    path = prefix_directory_join(prefix, f"{prefix}_synonyms.tsv")
    header = [f'{prefix}_id', f'synonym']

    @cached_multidict(path=path, header=header)
    def _get_multidict() -> Mapping[str, List[str]]:
        graph = get_obo_graph(prefix, url=url)
        rv = multidict(_iterate(graph, prefix))
        return {
            k: sorted(set(v))
            for k, v in rv.items()
        }

    return _get_multidict()


def _iterate(graph: nx.MultiDiGraph, prefix: str) -> Iterable[Tuple[str, str]]:
    for node, data in graph.nodes(data=True):
        if not node.lower().startswith(f'{prefix.lower()}:'):
            continue

        name = data['name']
        yield node, name

        for synonym in data.get('synonym', []):
            synonym = synonym.strip('"')
            if not synonym:
                continue

            if "RELATED" in synonym:
                synonym = synonym[:synonym.index('RELATED')].rstrip().rstrip('"')
            elif "EXACT" in synonym:
                synonym = synonym[:synonym.index('EXACT')].rstrip().rstrip('"')
            elif "BROAD" in synonym:
                synonym = synonym[:synonym.index('BROAD')].rstrip().rstrip('"')
            else:
                logger.warning(f'For {node} unhandled synonym: {synonym}')
                continue

            yield node, synonym



# -*- coding: utf-8 -*-

"""Converter for MGI."""

from collections import defaultdict
from typing import Iterable

import pandas as pd
from tqdm import tqdm

from ..path_utils import ensure_df
from ..struct import Obo, Reference, Synonym, Term, from_species

PREFIX = 'mgi'
MARKERS_URL = 'http://www.informatics.jax.org/downloads/reports/MRK_List2.rpt'
ENTREZ_XREFS_URL = 'http://www.informatics.jax.org/downloads/reports/MGI_EntrezGene.rpt'
ENSEMBL_XREFS_URL = 'http://www.informatics.jax.org/downloads/reports/MRK_ENSEMBL.rpt'


def get_obo() -> Obo:
    """Get MGI as OBO."""
    terms = list(get_terms())
    return Obo(
        ontology=PREFIX,
        name='Mouse Genome Database',
        terms=terms,
        typedefs=[from_species],
        auto_generated_by=f'bio2obo:{PREFIX}',
    )


COLUMNS = ['MGI Accession ID', 'Marker Symbol', 'Marker Name']


def get_ensembl_df() -> pd.DataFrame:
    """Get the Ensembl mappings dataframe."""
    return ensure_df(
        PREFIX,
        ENSEMBL_XREFS_URL,
        sep='\t',
        names=[
            'mgi_id',
            'symbol',
            'name',
            'position',
            'chromosome',
            'ensembl_accession_id',
            'ensembl_transcript_id',
            'ensembl_protein_id',
            'feature_type',
            'start',
            'end',
            'strand',
            'biotypes',
        ],
    )


def get_entrez_df() -> pd.DataFrame:
    """Get the Entrez mappings dataframe."""
    return ensure_df(
        PREFIX,
        ENTREZ_XREFS_URL,
        sep='\t',
        names=[
            'mgi_id',
            'symbol',
            'status',
            'name',
            'position',
            'chromosome',
            'Type',
            'Secondary Accession IDs',
            'entrez_id',
            'synonyms',
            'Features',
            'start',
            'end',
            'strand',
            'biotypes',
        ],
        dtype={
            'entrez_id': str,
        }
    )


def get_terms() -> Iterable[Term]:
    """Get the MGI terms."""
    df = ensure_df(PREFIX, MARKERS_URL, sep='\t')

    entrez_df = get_entrez_df()
    mgi_to_entrez_id, mgi_to_synonyms = {}, {}
    for mgi_id, synonyms, entrez_id in entrez_df[['mgi_id', 'synonyms', 'entrez_id']].values:
        mgi_id = mgi_id[len('MGI:'):]
        if synonyms and pd.notna(synonyms):
            mgi_to_synonyms[mgi_id] = synonyms.split('|')
        if entrez_id and pd.notna(entrez_id):
            mgi_to_entrez_id[mgi_id] = entrez_id

    ensembl_df = get_ensembl_df()
    # ensembl_df.to_csv('test.tsv', sep='\t', index=False)
    mgi_to_ensemble_ids = defaultdict(list)
    ensembl_columns = ['mgi_id', 'ensembl_accession_id', 'ensembl_transcript_id', 'ensembl_protein_id', ]
    ensembl_it = ensembl_df[ensembl_columns].values
    for mgi_id, ensemble_accession_id, ensemble_transcript_ids, ensemble_protein_ids in ensembl_it:
        mgi_id = mgi_id[len('MGI:'):]
        if ensemble_accession_id and pd.notna(ensemble_accession_id):
            mgi_to_ensemble_ids[mgi_id].append(ensemble_accession_id)
        if ensemble_transcript_ids and pd.notna(ensemble_transcript_ids):
            for ensemble_transcript_id in ensemble_transcript_ids.split():
                mgi_to_ensemble_ids[mgi_id].append(ensemble_transcript_id)
        if ensemble_protein_ids and pd.notna(ensemble_protein_ids):
            for ensemble_protein_id in ensemble_protein_ids.split():
                mgi_to_ensemble_ids[mgi_id].append(ensemble_protein_id)

    for identifier, name, definition in tqdm(df[COLUMNS].values, total=len(df.index), desc=f'Mapping {PREFIX}'):
        identifier = identifier[len('MGI:'):]

        synonyms = []
        if identifier in mgi_to_synonyms:
            for synonym in mgi_to_synonyms[identifier]:
                synonyms.append(Synonym(name=synonym))

        xrefs = []
        if identifier in mgi_to_entrez_id:
            xrefs.append(Reference(prefix='ncbigene', identifier=mgi_to_entrez_id[identifier]))
        if identifier in mgi_to_ensemble_ids:
            for ensembl_id in mgi_to_ensemble_ids[identifier]:
                xrefs.append(Reference(prefix='ensembl', identifier=ensembl_id))

        term = Term(
            reference=Reference(prefix=PREFIX, identifier=identifier),
            name=name,
            definition=definition,
            xrefs=xrefs,
            synonyms=synonyms,
        )
        term.append_relationship(
            from_species,
            Reference(prefix='taxonomy', identifier='10090', name='Mus musculus'),
        )
        yield term


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Converter for CGNC."""

from typing import Iterable

import pandas as pd

from ..path_utils import ensure_path
from ..struct import Obo, Reference, Synonym, Term, from_species

PREFIX = 'cgnc'
URL = "http://birdgenenames.org/cgnc/downloads.jsp?file=standard"


def get_obo() -> Obo:
    """Get CGNC as OBO."""
    terms = list(get_terms())
    return Obo(
        terms=terms,
        name='CGNC',
        ontology=PREFIX,
    )


def get_terms() -> Iterable[Term]:
    """Get CGNC terms."""
    path = ensure_path(PREFIX, URL, path=f'{PREFIX}.tsv')
    df = pd.read_csv(path, sep='\t', dtype={'Entrez Gene id': str, 'CGNC id': str})

    for cgnc_id, entrez_id, ensembl_id, symbol, name, synonyms, status, _ in df.values:
        xrefs = []
        if entrez_id and pd.notna(entrez_id):
            xrefs.append(Reference(prefix='ncbigene', identifier=entrez_id))
        if ensembl_id and pd.notna(ensembl_id):
            xrefs.append(Reference(prefix='ensembl', identifier=ensembl_id))

        if synonyms and pd.notna(synonyms):
            synonyms = [
                Synonym(name=synonym)
                for synonym in synonyms.split('|')
            ]
        else:
            synonyms = []

        term = Term(
            name=symbol,
            reference=Reference(prefix=PREFIX, identifier=cgnc_id),
            xrefs=xrefs,
            synonyms=synonyms,
            definition=name,
        )
        term.append_relationship(from_species, Reference(
            prefix='taxonomy', identifier='9031', name='Gallus gallus',
        ))
        yield term


if __name__ == '__main__':
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Data structures for registries."""

from dataclasses import dataclass

__all__ = [
    'Registry',
    'miriam',
    'ols',
    'obofoundry',
]


@dataclass
class Registry:
    """Represents a place to look up namespaces."""

    #: The name of the registry
    name: str

    #: The url that can be used to look up a resource.
    resolution_url: str

    def resolve_resource(self, resource: str) -> str:
        """Get the URl for the given resource."""
        return self.resolution_url + resource


miriam = Registry(
    name='miriam',
    resolution_url='https://registry.identifiers.org/registry/'
)

ols = Registry(
    name='ols',
    resolution_url='https://www.ebi.ac.uk/ols/ontologies/'
)

obofoundry = Registry(
    name='obofoundry',
    resolution_url='http://www.obofoundry.org/ontology/',
)



# -*- coding: utf-8 -*-

"""Data structures for OBO."""

from .parser import get_terms_from_graph  # noqa: F401
from .struct import Obo, Reference, Synonym, SynonymTypeDef, Term, TypeDef, from_species  # noqa: F401
